#########################################################
#########################################################
####    This file contains the core model logic      ####
####          for the hypercycles3 model.            ####
#########################################################
#########################################################
# This work is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License.
# http://creativecommons.org/licenses/by-nc/3.0/



## Absolutely required packages:
library(igraph)
library(sna,warn.conflicts=FALSE)

## Create a default parameter list.
## This function generates a generic default argument
## list that can be specified at call time or edited
## after the fact.
hyper3params <- function(
    n=100,                  ## sqrt(n) x sqrt(n) grid if initialEdges=="moore"
    maxBallType=3,          
    maxBallsInSystem=200,   
    maxEdgesInSystem=800,   
    numRulesEachType=20,
    initialEdges="moore",   ## c("moore","random","full")
    fixedOutDegree=FALSE,   ## should network out-degree be fixed at init values?
    allowSelfPass=FALSE,    
    killEmptyBins=FALSE,    
    ruleFlux=0,             ## 0<=x<=1, chance at each tick of random rule appearing
    addrFlux=0,             ## 0<=x<=1, chance at each tick of random addres appearing
    deathRate=Inf,          ## >=1 one cell dies every deathRate ticks
    replaceWithRandom=TRUE, ## should dead inventory be randomized for redist?
    diffusionRate=0,
    inheritedMean=.7,       ## mean proportion of inventory to pass on [0,1]
    inheritedVar=.001,         ## variance of proportion of inv to pass on (>0)
    reproduction="target",  ## c("source","target","reciprocal")
    allowBrokerage=FALSE,    
    percentAddrReprod=0,    ## 0<=x<=1
    brokerageReliability=1, ## how likely is brokerage to succeed at each step?
    chemistry="all",        ## c("all","solo")
    searchType="random",    ## c("random","selective")
    env.endogenous=FALSE,    
    env.rich=TRUE,
    env.startingBalls=200   ## ignored if !env.endogenous
){
    return(as.list(sys.frame(sys.nframe())))
}



###########################
## newWorld()
##  Creates a new 'world' object based on the given parameters.
##  This object is a classed list() with parameter values and other variables
##  describing the state of the entire simulation environment, bins and
##  inventories included.
##  After creation, functions like iterWorld() and (more commonly) runWorld()
##  are used to push the world through ticks.
###########################

newWorld <- function(params=hyper3params()){
    
    ## all the work is done within parameter list, so 
    ## make sure it's a list and classed right
    params <- as.list(params)
    class(params) <- c("hyper3world","list")
    within(params,{ ## drop inside of parameter list
    
    
        ############
        ## tick: current time step
        ############
        tick <- 0 ## brand new world
        
        
        ############
        ## set up parameters for inheritance of traits
        ############
        
        ## first, decide if proportion of inherited inventory is fixed
        ## or stochastic. (That is, test if inheritedVar is positive)
        if(!(stochasticInheritance <- inheritedVar > 0)){
            ## if inherited proportion is not stochastic, still set
            ## inheritedVar to a very small number. This allows things
            ## to function right without resorting to cases later on.
            inheritedVar <- 1e-20
        }
        
        ## now derive the parameters for stochastic inheritance.
        ## these convert a mean and variance into alpha and beta parameters
        ## for a Beta distribution.
        betaA <- getBetaA(inheritedMean,inheritedVar)
        betaB <- getBetaB(inheritedMean,inheritedVar)
        ## a few checks on these parameters
        if(betaA<1 || betaB<1) warning(
            "Multimodal distribution for inheritance. (Try a smaller variance?)"
        )
        if(is.nan(betaA) || is.nan(betaB)) stop(
            "NaN Beta parameters. Probably too large a value for inheritedVar."
        )
        
        
        
        
        ############
        ## initialize up address inventories. These are put into an
        ## adjacency matrix of counts.
        ############

        ## if "moore", make sure n is a perfect square, otherwise override
        if(initialEdges=="moore" && sqrt(n)%%1>0){
            warning("n not a perfect square, setting initialEdges='random'")
            initialEdges <- "random"
        }else if(initialEdges=="full"){ 
            ## for "full" graph make sure all the edges are allowed
            maxEdgesInSystem <- n^2
        }
        
        ## if maxEdgesInSystem is NULL, then just set it to the number of edges
        ## that would exist in a moore-graph (even if n can't be moore).
        if(is.null(maxEdgesInSystem)){
            maxEdgesInSystem <- n*8
        }
        
        ## 'am' holds the address inventory matrix.
        ## if initialEdges=='moore', give everybody one address to each of
        ## their Moore neighbors. If initialEdges=='random', then distribute
        ## maxEdgesInSystem edges uniformly and independently among possible
        ## edges (multiples allowed). If initialEdges=='full', then every 
        ## possible edge is realized.
        am <- switch(initialEdges,
            moore = get.adjacency(agraph.moore(n)),
            random = array(distribute(maxEdgesInSystem,n^2,random=TRUE),dim=c(n,n)),
            full = matrix(1,nrow=n,ncol=n)
        )
        
        ## add all self-edges if 'moore' or 'full' and allowSelfPass
        if(allowSelfPass && (initialEdges == "moore" || initialEdges =="full")){
            diag(am) <- 1
        }
        
        ## record the current number of edges
        numEdges <- sum(am)
        
        ## the above procedure for 'random' edge distribution allows 
        ## self-addresses. This loop makes sure those edges are randomly
        ## redistributed if self-passing is turned off.
        while(!allowSelfPass && any(diag(am)>0)){ # inefficient, but just once
            diag(am) <- 0
            am <- am + distribute(numEdges-sum(am),n^2)
        }
        
        
        
        ############
        ## set up ballEnv, the environment from which bins draw balls .
        ## ballEnv is a numeric vector, the i'th element of which
        ## is the count (possible Inf) of the i'th ball.
        ############
        
        ## first set ballEnv assuming endogenous environment 
        ballEnv <- 
            if(env.rich){
                ## if rich environment, all counts are same
                rep(env.startingBalls,maxBallType)
            }else{
                ## otherwise first count is positive, rest are zero
                c(env.startingBalls,rep(0,maxBallType-1))
            }
        ## now if environment is not endogenous, set any postive counts
        ## to Inf
        if(!env.endogenous){
            ballEnv[ballEnv>0] <- Inf
        }
        
        ## one last detail.
        ## infinite values don't work in `prob` argument to sample, so we make this:
        probBallEnv <- if(env.endogenous){
            ballEnv
        }else{
            is.infinite(ballEnv)
        }
        
        
        
        
        
        ############
        ## Create rule types based on chemistry. This is a numRules x 2
        ## matrix giving input ball and output ball labels. Inventories are
        ## handled next.
        ############
        
        if(chemistry=="all"){ # 'all' chemistry
            ## combn() gives all (i,j) combos where i<j (half of what we want),
            ## so we combine its ouput with its mirror image to get all rules.
            ruleList <- t(combn(maxBallType,2))
            ruleList <- rbind(ruleList,ruleList[,2:1])
            colnames(ruleList) <- c("inBall","outBall") ## be sure to name cols
        }else if(chemistry=="solo"){ # 'solo' chemistry
            ## solo rule list is simple to construct
            ruleList <- cbind(inBall=1:maxBallType,outBall=c(2:maxBallType,1))
        }else{ # catch for misspecified chemistry
            stop("don't know what to do with chemistry '",chemistry,"'.")
        }
        
        
        
        ############
        ## Make a ball inventory for each bin.
        ## Represented by an n-by-maxBallType matrix with counts.
        ############
        
        ## start out empty
        ballInv <- matrix(0,nrow=n,ncol=maxBallType)
        
        ## For each ball type, distribute a number of balls to the bins
        ## uniformly and independently (multiples allowed).
        ## Number of balls distributed is maxBallsInSystem/maxBallType,
        ## rounded down. (thus initial balls in system may be less than
        ## maxBallsInSystem).
        for(i in seq.int(maxBallType)){
            ballInv[,i] <- 
                distribute(floor(maxBallsInSystem/maxBallType),n,TRUE)
        }
        
        
        ############
        ## Make rule inventory for each bin.
        ## Represented by an n-by-numRules matrix with counts
        ############
        
        ## start out empty
        ruleInv <- matrix(0,nrow=n,ncol=nrow(ruleList))

        ## For each rule type, distribute a number of rules to the bins
        ## uniformly and independently (multiples allowed).
        ## Number of rules distributed is numRulesEachType.
        for(i in seq.int(nrow(ruleList))){
            ruleInv[,i] <- distribute(numRulesEachType,n,TRUE)
        }
        
        
        ############
        ## Create a logical vector indicating which bins are living.
        ## 'Living' is defined here as having at least one out-neighbor
        ## (ie at least one address) _or_ at least one rule.
        ############
        
        isAlive <- rowSums(am) | rowSums(ruleInv)
        
        
        
        ############
        ## These are utility variables to make reproduction easier later.
        ## Both get set to TRUE if reproduction is "reciprocal".
        ############
        reproduceSource <- reproduction %in% c("source","reciprocal")
        reproduceTarget <- reproduction %in% c("target","reciprocal")
    })
}


###########################
## A 'print' and 'summary' method for 'hyper3world' objects:
###########################
print.hyper3world <- function(w){
    cat("hyper3 'world' object\n")
}
summary.hyper3world <- function(w){
    cat("Parameters:\n")
    print(as.data.frame(w[c('maxBallType','chemistry','reproduction','searchType','env.endogenous','env.rich')]))
    if(w$allowBrokerage){
        cat("\nbrokering with reliability", round(w$brokerageReliability,2),",\n")
        cat("reproducing",round(w$percentAddrReprod,3),"of addresses.\n")
    }else{
        cat("\n(unbrokered)\n")
    }
    cat("\n",if(worldIsAlive(w)){"Alive at tick "}else{"Dead at tick "},w$tick,".\n",sep="")
}



###########################
## iterWorld()
##  This is the bulk of the model logic. It takes a 'world' list (as output
##  by newWorld, iterWorld, runWorld, etc), and runs it through niter ticks.
##  It is designed for speed rather than feedback, so most printing/plotting
##  is done through runWorld function below.
###########################

iterWorld <- function(world,niter=1){
    ############
    ## For convenience it is allowed to feed just a parameter list
    ## to iterWorld, which it will then convert to a full world using
    ## newWorld.
    ############
    
    if(is.null(world$ballInv)){
        world <- newWorld(world)
    }
    
    
    
    
    ## because the work is done inside of the world list, we push niter inside.
    world$niter <- niter
    
    
    
    
    ############
    ## Do all the work inside the 'world' list, immediately returning the
    ## result.
    ############
    
    within(world,{
        for(thisIter in seq.int(niter)){ ## do all of this niter times
            ## increment tick
            tick <- tick+1
            
            
            ############
            ## Flux. Do random rules or addresses pop into existence?
            ############
            
            ## first rule flux
            if(ruleFlux>0 && runif(1)<ruleFlux){
                ## somebody gets a random rule!
                ## (pick an index and increment)
                thisIndex <- sample(length(ruleInv),1)
                ruleInv[thisIndex] <- ruleInv[thisIndex] + 1
                ## and somebody loses a rule.
                ## (pick an index and decrement)
                thisIndex <- sample(length(ruleInv),1,prob=ruleInv)
                ruleInv[thisIndex] <- ruleInv[thisIndex] - 1
            }
            
            ## then address flux
            if(addrFlux>0 && runif(1)<addrFlux){
                ## somebody gets a random address!
                ## (pick an index and increment)
                thisIndex <- sample(length(am),1)
                am[thisIndex] <- am[thisIndex] + 1
                
                ## and somebody loses an address.
                ## (pick an index and decrement)
                if(fixedOutDegree){ ## must decrement from this row
                    thisRow <- row(am)[thisIndex]
                    thisCol <- sample(length(am[thisRow,]),1,prob=am[thisRow,])
                    am[thisRow,thisCol] <- am[thisRow,thisCol] - 1
                }else{ ## decrement anybody you want
                    thisIndex <- sample(length(am),1,prob=am)
                    am[thisIndex] <- am[thisIndex] - 1
                }
            }
            
            
            
            ############
            ## Set curBall to NULL at the beginning of each tick
            ## (ball 'in hand')
            ############
            curBall <- NULL
            
            
            ############
            ## pick a bin and a rule randomly, weighted by rule count.
            ## Do this by picking a single (row,col) from ruleInv, weighted
            ## by the count in that (row,col). Then translate this to a bin
            ## number and a rule number.
            ############
            
            thisInd <- sample(length(ruleInv),1,prob=ruleInv)
            thisBin <- row(ruleInv)[thisInd]
            thisRule <- col(ruleInv)[thisInd]
            
            
            
            ############
            ## Try to get thisRule's input ball from environment
            ############
            
            
            ## leave curBall as NULL *unless* the following statement holds
            if(
                (
                    searchType=="selective" &&      ## (selective search AND available)
                    ballEnv[ruleList[thisRule,"inBall"]]>0
                ) ||                                ## OR
                (
                    searchType=="random" &&         ## (random search AND lucky)
                    sample(maxBallType,1,prob=probBallEnv) == ruleList[thisRule,"inBall"]
                )
            ){
                ## take the ball 'in hand' (curBall)
                curBall <- ruleList[thisRule,"inBall"]
                if(env.endogenous){
                    # remove from ballEnv by decrementing count
                    ballEnv[curBall] <- ballEnv[curBall]-1
                    ## update probBallEnv
                    probBallEnv <- ballEnv
                }
            }
            
            
            
            
            ############
            ## As long as we have a ball (curBall is not NULL),
            ## keep passing it around.
            ############
            
            while(!is.null(curBall)){
                ############
                ## First convert the ball according to the rule.
                ############
                
                curBall <- ruleList[thisRule,"outBall"]
                
                
                
                ############
                ## If thisBin has no neighbors, throw the converted ball
                ## ball back (if endogenous env), otherwise pass it on.
                ############
                
                if(!any(am[thisBin,]>0)){ ## thisBin has no neighbors
                    ## throw it back if appropriate
                    if(env.endogenous){
                        ballEnv[curBall] <- ballEnv[curBall] + 1
                        probBallEnv <- ballEnv
                    }
                    ## set curBall to NULL to stop the loop
                    curBall <- NULL
                }else{
                    ############
                    ## thisBin has at least one neighbor, so we try to make
                    ## a pass.
                    ############
                    
                    ## which rules are compatible with curBall as input?
                    compatRules <- which(ruleList[,"inBall"]==curBall)
                    
                    ############
                    ## Pick a neighbor, weighted by address inventory.
                    ## This starts the 'chain' vector, which can grow
                    ## in the case of brokerage.
                    ##  (chain starts with first recipient and ends with 
                    ##  whoever will get the ball; unbrokered chain has length 1)
                    ############
                    chain <- sample(n,1,prob=am[thisBin,])

                    
                    
                    ############
                    ## If brokerage turned on, keep appending 'chain' until
                    ## we find a compatible rule or there's nowhere left to go.
                    ############
                    if(allowBrokerage){
                        while(!any(ruleInv[tails(chain),compatRules]>0)){
                            ############
                            ## Calculate each bin's probability based on number
                            ## of addresses from end of chain and loops in 
                            ## chain not allowed.
                            ############
                            theseProbs <- am[tails(chain),]
                            theseProbs[c(thisBin,chain)] <- 0 ## no loops


                            ############
                            ## See if the chain manages to continue, based on
                            ## eligible bin and "brokerage reliability" param.
                            ############
                            if( ## nowhere to pass or brokerage unreliable
                                !any(theseProbs>0) || (
                                    brokerageReliability<1 && ## (speedier)
                                    runif(1)>brokerageReliability
                                )
                            ){
                                ## if brokerage fails, throw it back
                                if(env.endogenous){
                                    ballEnv[curBall] <- ballEnv[curBall] + 1
                                    probBallEnv <- ballEnv
                                }
                                curBall <- NULL
                                break ## break from brokerage loop
                            }
                            
                            
                            ############
                            ## If we're still here then brokerage continues,
                            ## so pick a (weighted) eligible bin and tack it on.
                            ############
                            chain <- c(chain,sample(n,1,prob=theseProbs))
                        }
                    }else{ # no brokerage
                        ############
                        ## If brokerage is turned off, then we simply check
                        ## if the neighbor is compatible.
                        ############
                        if(!any(ruleInv[tails(chain),compatRules]>0)){
                            ## incompatible, toss into environment.
                            if(env.endogenous){
                                ballEnv[curBall] <- ballEnv[curBall] + 1
                                probBallEnv <- ballEnv
                            }
                            curBall <- NULL
                        }else{
                            ## leave it.
                            ## (just in case, check:)
                            if(length(chain)>1){
                                stop("brokerage chain too long for unbrokered",
                                    "model (",length(chain),")")
                            }
                        }
                    }
                    ############
                    ## This ends the brokerage code.
                    ############



                    ############
                    ## If curBall is still not NULL, then there was a successful pass
                    ## (either brokered or not). So go through and reproduce edges and rules
                    ## as appropriate, and kill off edges and rules as appropriate.
                    ############
                    nextBin <- tails(chain) ## this is the bin that contains the rule
                    if(!is.null(curBall) && thisBin != nextBin){
                        ############
                        ## First pick which of the possibly several compatible rules is the one 
                        ## that will be used.
                        ############
                        ## the following convoluted if-else just picks a compatible rule from nextBin
                        nextRule <- if(length(compatRules)==1){ 
                            compatRules[1]
                        }else{
                            sample(
                                seq.int(nrow(ruleList))[compatRules],
                                1,prob=ruleInv[nextBin,compatRules]
                            )
                        }



                        ############
                        ## Now reproduce and kill off Rules, as appropriate.
                        ############
                        if(reproduceTarget){ ## target or reciprocal reproduction
                            ## increment count of nextRule in nextBin
                            ruleInv[nextBin,nextRule] <- ruleInv[nextBin,nextRule] + 1
                            ## pick a random rule globally to kill
                            toDelete <- sample(length(ruleInv),1,prob=ruleInv)
                            ## decrement count of rule to kill
                            ruleInv[toDelete] <- ruleInv[toDelete] - 1
                        }
                        if(reproduceSource){ ## source or reciprocal reproduction
                            ## increment count of thisRule in thisBin
                            ruleInv[thisBin,thisRule] <- ruleInv[thisBin,thisRule] + 1
                            ## pick a random rule globally to kill
                            ## decrement count of rule to kill
                            ruleInv[toDelete] <- ruleInv[toDelete] - 1
                        }



                        ############
                        ## Now create and kill off edges (addresses), as appropriate.
                        ## (this is more complicated with edges than it is with rules).
                        ## We first do all of the edge creation, then iterate edge death 
                        ## to bring the total edge count back down.
                        ############
                        ## First form of address creation is simple reproduction
                        ############
                        if(percentAddrReprod > 0){
                            reproduceMe <- (
                                ## the 'distribute' call below picks percentAddrReprod addresses used in
                                ##  the brokerage chain to be reproduced. The command rounds up, so to
                                ##  reproduce only one address, simply set percentAddrReprod to any number
                                ##  less than 1/n.
                                
                                #distribute(
                                #    ceiling((length(chain))*percentAddrReprod),
                                #    length(chain)
                                #) > 0
                                
                                ## the following line is an alternative to the above distribute() code.
                                ##  it will give each address independent uniform probability of reproducing:
                                
                                runif(length(chain))<percentAddrReprod
                            )
                            
                            ## now increment every address chosen above, and decrement if fixedOutDegree
                            for(i in which(reproduceMe)){
                                thisRow <- c(thisRule,chain)[i]
                                am[thisRow,chain[i]] <- am[thisRow,chain[i]] + 1
                                if(fixedOutDegree){ ## delete one edge from this row
                                    thisCol <- sample(am[thisRow,],1,prob=am[thisRow,])
                                    am[thisRow,thisCol] <- am[thisRow,thisCol] - 1
                                } ## else case taken care of in bulk deletions later on
                            }
                        }
                        
                        ############
                        ## Next we create new edges through diffusion.
                        ############
                        if(runif(1) < diffusionRate){
                            ## pick a teacher and a lesson
                            toTeach <- which(rmultinom(1,1,prob=c(
                                sum(am[thisRule,]),
                                sum(ruleInv[thisRule,]),
                                sum(am[nextRule,]),
                                sum(ruleInv[nextRule,])
                            ))>0)
                            if(toTeach==1){ ## source teaches address
                                i <- sample(n,1,prob=am[thisRule,])
                                am[nextRule,i] <- am[nextRule,i] + 1
                                if(fixedOutDegree){ ## delete one edge from this row
                                    thisCol <- sample(am[nextRule,],1,prob=am[nextRule,])
                                    am[nextRule,thisCol] <- am[nextRule,thisCol] - 1
                                } ## else case taken care of in bulk deletions later on
                            }else if(toTeach==2){ ## source teaches rule
                                i <- sample(
                                    nrow(ruleList),1,
                                    prob=ruleInv[thisRule,]
                                )
                                ruleInv[nextRule,i] <- ruleInv[nextRule,i] + 1
                            }else if(toTeach==3){ ## target teaches address
                                i <- sample(n,1,prob=am[nextRule,])
                                am[thisRule,i] <- am[thisRule,i] + 1
                                if(fixedOutDegree){ ## delete one edge from this row
                                    thisCol <- sample(am[thisRule,],1,prob=am[thisRule,])
                                    am[thisRule,thisCol] <- am[thisRule,thisCol] - 1
                                } ## else case taken care of in bulk deletions later on
                            }else if(toTeach==4){ ## target teaches rule
                                i <- sample(
                                    nrow(ruleList),1,
                                    prob=ruleInv[nextRule,]
                                )
                                ruleInv[thisRule,i] <- ruleInv[thisRule,i] + 1
                            }
                        }
                        
                        ############
                        ## This loop to bring edge count back to where it should be.
                        ## (should only occur in !fixedOutDegree case)
                        ############
                        surplusEdges <- sum(am)-numEdges
                        remainder <- surplusEdges %% 1 ## in case of non-integer edge counts (shouldn't happen)
                        while(surplusEdges>0){
                            # delete addresses to numEdges
                            toDelete <- sample(n^2,1,prob=am)
                            if(am[toDelete]>=1){
                                del <- min(1,surplusEdges)
                                am[toDelete] <- am[toDelete] - del
                                surplusEdges <- surplusEdges - del
                                remainder <- surplusEdges %% 1
                            }else{
                                del <- am[toDelete]
                                am[toDelete] <- 0
                                surplusEdges <- surplusEdges - del
                                remainder <- surplusEdges %% 1
                            }
                        }
                        #if(any(am<1 & am>0)) recover() ## this will catch certain bad edge behavior.
                        
                        
                        
                        ############
                        ## Everything is complete for this pass, just need to prepare for the next pass.
                        ############
                        thisBin <- nextBin
                        thisRule <- nextRule
                    }
                }
            }
            
            
            
            
            
            
            ############
            ## Death and mutation.
            ## this section covers random cell death and distribution/mutation of its inventory
            ############
            ## First update the isAlive logical vector
            ## (bin must have at least one out-address or at least one rule in its inventory)
            isAlive <- rowSums(am) | rowSums(ruleInv)
            
            ## Do all of this only if we are in one of the 'deathRate'th ticks. (abuse of language :)
            if(deathRate<Inf && tick %% deathRate == 0){
                ############
                ## Figure out which bin will die and how much of it will be passed on:
                ############
                
                ## first pick the bin that's going to die.
                binToDie <- sample(n,1,prob=isAlive)
                ## get the counts of its address and rule inventories
                invCount <- c(am[binToDie,],ruleInv[binToDie,])
                ## pick the proportion of its inventory that is going to die using Beta distribution
                ##  (if inheritedVar is set to 0 this returns a fixed value)
                propToDie <- 1 - rbetaMV(1,inheritedMean,inheritedVar)
                ## translate that to the number to die
                numToDie <- propToDie*sum(invCount)
                
                ############
                ## Iterate until there are is no more inventory left that needs to die
                ############
                while(numToDie>0){
                    ## thisInv picks one piece of inventory (address or rule)
                    thisInv <- sample(length(invCount),1,prob=invCount)
                    
                    ############
                    ## Execute seperate code if the item to be killed off is an address or a rule.
                    ############
                    if(thisInv<=n){ ## address
                        ############
                        ## First the code for address killing.
                        ############
                        ## decrement the address count
                        am[binToDie,thisInv] <- am[binToDie,thisInv] - 1
                        ############
                        ## If replaceWithRandom is turned on then increment a uniformly random address
                        ## somewhere in the system (except self-loops).
                        ## If replaceWithRandom is turned off then transfer the address to a uniformly random
                        ## node (unless that makes a loop).
                        ############
                        if(replaceWithRandom){
                            thisI <- sample(n,1) ## pick a 'from'
                            thisJ <- sample(seq.int(n)[-thisI],1) ## pick a 'to'
                            am[thisI,thisJ] <- am[thisI,thisJ] + 1 ## increment
                            if(fixedOutDegree){ ## delete one edge from this row
                                thisJ <- sample(am[thisI,],1,prob=am[thisI,])
                                am[thisI,thisJ] <- am[thisI,thisJ] - 1
                            } ## else case taken care of in bulk deletions later on
                        }else{
                            thisI<- sample(seq.int(n)[-thisInv],1) ## pick a 'from'
                            am[thisI,thisInv] <- am[thisI,thisInv] + 1 ## increment
                            if(fixedOutDegree){ ## delete one edge from this row
                                thisJ <- sample(am[thisI,],1,prob=am[thisI,])
                                am[thisI,thisJ] <- am[thisI,thisJ] - 1
                            } ## else case taken care of in bulk deletions later on
                        }
                    }else if(numToDie>=1){ # rule
                        ############
                        ## Now the code for rule killing.
                        ############
                        thisInv <- thisInv-n ## to fix the index
                        ## decrement the rule count
                        ruleInv[binToDie,thisInv] <- ruleInv[binToDie,thisInv]-1
                        ############
                        ## If replaceWithRandom is turned on then increment a uniformly random rule
                        ## somewhere in the system.
                        ## If replaceWithRandom is turned off then transfer the rule to a uniformly random
                        ## node.
                        ############
                        if(replaceWithRandom){
                            thisI <- sample(n,1) ## pick a bin
                            thisJ <- sample(nrow(ruleList),1) ## pick a rule
                            ruleInv[thisI,thisJ] <- ruleInv[thisI,thisJ] + 1 ## increment
                        }else{
                            thisI <- sample(n,1) ## pick a bin
                            ruleInv[thisI,thisInv] <- ruleInv[thisI,thisInv] + 1 ## increment
                        }
                    }
                    ## the following is a catch-all to avoid certain infinite loops if non-integer
                    ## inventories are allowed.
                    if(all(am[binToDie]==0) && numToDie<1){
                        break
                    }
                    
                    ############
                    ## Unpdate the inventory count and numToDie.
                    ############
                    invCount <- c(am[binToDie,],ruleInv[binToDie,])
                    numToDie <- propToDie*sum(invCount)
                }
            }
            
            ############
            ## This code is primarily for hyper2 compatibility. Used to be more involved,
            ## but now it just indicates the dead bins if killEmptyBins is turned on.
            ## (used to actually delete them, too, but that ended up being unnecessary)
            ############
            if(killEmptyBins){
                ## this just means we delete addresses to and from any bins
                ## with no rules.
                ## actually, deleting those addresses makes for infinite loops
                ## ( eventually no way for chains to die )##
                deadBins <- rowSums(ruleInv)==0
                #am[deadBins,] <- 0
                #am[,deadBins] <- 0
                #numRules <- sum(am)
            }
            
            

        } ## end giant for loop
        
        
        
        ## checks for simple errors that might be introduced in iterations 
        stopifnot(
            all(ballEnv>=0), ## no negative ball counts
            all(ruleInv>=0), ## no negative rule inventory
            !env.endogenous || all(probBallEnv==ballEnv), ## matching probBallEnv
            env.endogenous || all(ballEnv %in% c(0,Inf)) ## endogenous means 0 or Inf
        )
    }) ## end within()
}




###########################
## runWorld()
##  This is the common way to actually run the model. It breaks up `niter`
##  into chunks of size `displayEvery` and calls iterWorld on those chunks.
##  Between every call to iterWorld some output is produced and the world
##  is checked for death.
###########################

runWorld <- function(
    world,niter=NULL,
    displayEvery=100,
    stopOnDeath=TRUE,prefix=""
){
    startingTick <- world$tick
    if(is.null(niter)){ ## if niter not supplied
        niter <- world$niter
    }#else{
    #    world$niter <- niter
    #}
    if(is.null(niter)){ ## if niter still not there
        stop("need niter argument or niter element in world")
    }
    niter.startingVal <- niter

    thisIter <- 0
    while(thisIter<niter){
        catMe <- prefix
        cat(catMe,"tick:",world$tick,"\n",sep=" ")
        thisIter <- thisIter + displayEvery
        world <- iterWorld(world,min(displayEvery,startingTick + niter - world$tick))
    }
    world$cm <- getCompatibilityMatrix(world)
    world$bcm <- getBinCompatibilityMatrix(world)
    cat(prefix,"tick:",world$tick,"\n",sep=" ")
    return(world)
}




###########################
## Supplementary model logic
##  The following functions implement supplementary model logic, useful
##  both while running the model and in analysis.
###########################


## Creates a graph from living rule-bins, directed edges defined
## by rule- and address-compatibility.
## (Returns NA for too dense a network)
getCompatibilityMatrix <- function(w,cmSizeCutoff=40){
    cm <- w$am ## start with the whole adjacency matrix
    rownames(cm) <- colnames(cm) <- seq.int(nrow(cm)) ## give each row a name
    
    binRuleCounts <- rowSums(w$ruleInv>0) ## how many rules living in each bin?
    ## new indices, one node for each rulexbin combo with at least one alive
    newInd <- unlist(mapply(rep,x=seq.int(nrow(cm)),times=binRuleCounts))
    
    ## after this, cm is just a rulexbin adjacency matrix, ignoring
    ## compatibility (edges based on bin adjacency only)
    cm <- cm[newInd,newInd]
    
    ## see if it's just way too big
    if(length(cm)>1 && nrow(cm)>cmSizeCutoff) return(NA)
    
    ## negate rulexbin edges with incompatible rules
    
    ## for each i,j combo that 'makes sense'
    for(i in unique(as.numeric(rownames(cm)))){
    for(j in which(w$am[i,]>0 & binRuleCounts>0)){
        ## for each i -> j rule combo check compatibility
        for(ii in seq.int(length(iRules <- which(w$ruleInv[i,]>0)))){
        for(jj in seq.int(length(jRules <- which(w$ruleInv[j,]>0)))){
            if(
                w$ruleList[iRules[ii],"outBall"] !=
                w$ruleList[jRules[jj],"inBall"]
            ){
                cm[
                    which(rownames(cm)==as.character(i))[ii],
                    which(colnames(cm)==as.character(j))[jj]
                ] <- 0
            }
        }}
    }}
    
    return(cm)
}

## translates output of getCompatibiltyMatrix by collapsing rules that share
## a bin and optionally filling out empty bins
getBinCompatibilityMatrix <- function(w,with.isolates=TRUE,binarize=TRUE){
    ## circumvent compatibility matrix if possible:
    if(binarize){
        bcm <- getBasicBinCompatibilityMatrix(w)
        if(!with.isolates){
            keepMe <- apply(bcm>0,1,any) | apply(bcm>0,2,any)
            bcm <- bcm[keepMe,keepMe]
        }
    }
    
    if(is.null(w$cm)){
        w$cm <- getCompatibilityMatrix(w)
    }
    
    ## first a degenerate case
    if(length(w$cm)<=1){ ## empty or singular compatibility matrix
        return(matrix(0,nrow=nrow(w$am),ncol=ncol(w$am)))
    }
    ## and another
    if(length(unique(colnames(w$cm)))<=1){
        return(matrix(0,nrow=nrow(w$am),ncol=ncol(w$am)))
    }
    
    ## now collapse w$cm
    dupedNames <- duplicated(rownames(w$cm))
    multiBins <- unique(rownames(w$cm)[dupedNames])
    bcm <- w$cm
    for(b in multiBins){
        bcm[b,] <- colSums(w$cm[rownames(w$cm)==b,])
        bcm[,b] <- rowSums(w$cm[,colnames(w$cm)==b])
    }
    bcm <- bcm[!dupedNames,!dupedNames]
    
    if(with.isolates){
        emptyBins <- seq.int(w$n)[-as.numeric(rownames(bcm))]
        ## add missing columns and give names
        bcm <- cbind(bcm,matrix(0,nrow=nrow(bcm),ncol=length(emptyBins)))
        colnames(bcm) <- c(rownames(bcm),emptyBins)
        ## add missing rows and give names
        bcm <- rbind(bcm,matrix(0,nrow=length(emptyBins),ncol=ncol(bcm)))
        rownames(bcm) <- colnames(bcm)
        ## reorder for consistency
        bcm <- bcm[order(as.numeric(rownames(bcm))),order(as.numeric(colnames(bcm)))]
    }
    
    
    if(binarize){
        bcm[bcm>1] <- 1
    }
    
    return(bcm)
}

## this function makes a bin compatibility matrix if world is too complex
## for getCompatibilityMatrix. Assumes `with.isolates' and `binarize'.
getBasicBinCompatibilityMatrix <- function(w){
    ## start with adjacency matrix and remove incompatible eges
    bcm <- w$am
    bcm[bcm>0] <- unlist(lapply(which(bcm>0),function(x){
        from <- row(bcm)[x]
        to <- col(bcm)[x]
        return(any(
            w$ruleList[w$ruleInv[from,]>0,"outBall"] %in%
            w$ruleList[w$ruleInv[to,]>0,"inBall"]
        ))
    }))
    return(bcm)
}

## Does the world have any cycles?
hasCycle <- function(am){
    if(length(am)>1){
        if(!any(kcycle.census(am,3,tabulate.by.vertex=FALSE)$cycle.count>0)){
            return(
                any(kcycle.census(am,9,tabulate.by.vertex=FALSE)$cycle.count>0))
        }else{
            return(TRUE)
        }
    }else{
        return(FALSE)
    }
}

## Is the world 'alive'?
worldIsAlive <- function(w,removeSelfPass=TRUE){
    ## first check the input environment
    if(is.null(w$cm)){
        w$cm <- getCompatibilityMatrix(w)
    }
    if(w$tick==0) return(TRUE) ## heuristic, always call a fresh world alive
    if(length(w$cm)<=1) return(FALSE)
    if(removeSelfPass && !is.null(w$allowSelfPass) && w$allowSelfPass){
        ## if we're allowing self pass, don't count those edges in cycles
        for(i in rownames(w$cm)[duplicated(rownames(w$cm))]){
            w$cm[rownames(w$cm)==i,colnames(w$cm)==i]
        }
    }
    missingBalls <- w$ballEnv<1
    compatRules <- w$ruleList[,"inBall"] %in% which(!missingBalls)
    if(any(missingBalls)){
        if(sum(compatRules)==1){
            #if(!any(w$cm[sum(w$ballInv[,!missingBalls])>0,]>0)){
            if(!any(
                w$cm[w$ruleInv[as.numeric(rownames(w$cm)),compatRules]>0,]>0
            )){
                return(FALSE)
            }
        }else{
            if(!any(
                w$cm[
                    rowSums(w$ruleInv[as.numeric(rownames(w$cm)),
                    compatRules])>0,
                ]>0
            )){
                ## if none of the bins that have a rule compatible with some
                ## input ball have an ouput address, then we call the system dead.
                return(FALSE)
            }
        }
    }
    return(hasCycle(w$cm))
}

















##################################################
##                  #########                   ##
####        ##########################        ####
##################################################
##   Some generally useful utility functions    ##
##################################################
####        ##########################        ####
##                  #########                   ##
##################################################


## distribute k items over n slots, uniformly or randomly.
## if k not multiple of n, distributes remainder randomly (without repeats).
distribute <- function(k,n,random=FALSE){
    if(random){
        res <- as.numeric(rmultinom(1,k,rep(1,n)))
    }else{
        res <- rep(floor(k/n),n)
        extras <- sample(n,k-sum(res),replace=FALSE)
        res[extras] <- res[extras] + 1
    }
    return(res)
}


## makes an igraph object with edges determined by moore neighborhoods
agraph.moore <- function(n,nrow=sqrt(n),ncol=sqrt(n)){
    require(igraph)
    ## initialize an agent graph with moore-neighborhood connections
    if(round(nrow)*round(ncol) != n)
        stop("nrow and ncol do not match n. Either they are misspecified or n ",
             "is not a perfect square")
    res <- graph.empty(n)
    v <- seq.int(n)-1
    edgelist <- c(matrix(c(
        v,((v %% nrow)+1)%%nrow + floor(v/nrow)*nrow, #n
        v,((v %% nrow)-1)%%nrow + floor(v/nrow)*nrow, #s
        v,(v-ncol)%%n, #e
        v,(v+ncol)%%n, #w
        v,((v%%nrow)+1)%%nrow + ((floor(v/nrow)*nrow+nrow)%%n), #ne
        v,((v%%nrow)-1)%%nrow + ((floor(v/nrow)*nrow+nrow)%%n), #se
        v,((v%%nrow)+1)%%nrow + ((floor(v/nrow)*nrow-nrow)%%n), #nw
        v,((v%%nrow)-1)%%nrow + ((floor(v/nrow)*nrow-nrow)%%n) #sw
    ),nrow=16,byrow=T))
    res <- add.edges(res,edgelist)
    E(res)$weight <- 1
    res$layout <- cbind(
        c(col(matrix(seq.int(n),nrow=nrow))),
        c(row(matrix(seq.int(n),nrow=nrow)))
    )
    return(res)
}

tails <- function(x){
    x[length(x)]
}



additionalIter <- function(wl,niter){ # world list, niter
    for(i in 1:length(wl)){
        cat(
            "\nrunning",niter,"additional iterations on world",
            i,"of",length(wl),"\n\n"
        )
        wl[[i]] <- runWorld(
            wl[[i]],
            niter,
            displayEvery=10000,
            doPlot=FALSE,
            stopOnDeath=TRUE
        )
    }
    return(wl)
}



## this function takes a hyper2 world and converts it to a 
## _partial_ hyper3 world. So far just enough to make a 
## compatibility matrix and plot it. (am,ruleList,ruleInv,etc.)
hyper2to3 <- function(h2,keepAllData=FALSE){ # keepAllData to retain hyper2 data
    # we're assuming a square moore graph:
    n <- dim(h2$agentArray)[4]^2
    # result list is either h2 or an empty list, depending on keepAllData:
    h3 <- if(keepAllData){h2}else{list()}
    # adjacency matrix:
    h3$am <- get.adjacency(agraph.moore(n))
    # ruleList:
    h3$ruleList <- h2$agentArray[,c("inBall","outBall"),1,1]
    # ruleInv:
    h3$ruleInv <- h2$agentArray[,"count",,]
    dim(h3$ruleInv) <- c(dim(h3$ruleInv)[1],n)
    h3$ruleInv <- t(h3$ruleInv)
    # ballEnv:
    h3$ballEnv <- h2$ballEnv[,"count"]
    # isAlive
    h3$deadBins <- rowSums(h3$ruleInv)==0
    # add some paramters for good measure:
    h3 <- c(h3,h2$params[c(
        "maxBallType","chemistry","reproduction"
    )])
    h3$env.endogenous <- h2$params$endogenous
    h3$env.rich <- h2$params$resourceMode=="rich"
    ## hyper3 has no searchType, but good to keep this data around
    h3$searchType <- h2$params$searchType
    # return the result:
    gc() ## (getting some mallocs in certain cases)
    return(h3)
}





#############################################
## These functions map between alpha/beta and
## mean/variance parameterization for the Beta
## Distribution.
#############################################

getBetaA <- function(m,v){     ## mean, variance
    m <- 2-1/m ## for some reason this works.
    (1-m)/(v*(2-m)^3) - 1/(2-m)
}
getBetaB <- function(m,v){     ## mean, variance
    m <- 2-1/m ## for some reason this works.
    (1-m)*getBetaA(1/(2-m),v)
}
betaMean <- function(A,B){A/(A+B)}
betaVar <- function(A,B){
    (A*B)/((A+B)^2 * (A+B+1))
}
rbetaMV <- function(n,m,v,...){
    if(v==0){
        return(rep(m,n))
    }else if(v<0 || m<0 || m>1){
        stop("mean or variance out of range")
    }else if(
        (a <- getBetaA(m,v))>0 ||
        (getBetaB(m,v))>0
    ){ ## go ahead:
        b <- getBetaB(m,v)
        if(a<1 || b<1){
            warning("multimodal beta distribution.")
        }
        return(rbeta(n,a,b,...))
    }else{
        stop("bad parameters for beta distribution. Try a smaller variance.")
    }
}




###########################
## A couple of approaches to plotting hyper3world objects
###########################

plotWorldRules <- function(w,useGrid=FALSE,label.cex=.7,vertex.cex=10,...){
    if(is.null(w$cm)){
        w$cm <- getCompatibilityMatrix(w)
    }
    if(is.na(w$cm)){
        stop('too big to plot')
    }
    if(is.null(w$bcm)){
        w$bcm <- getBinCompatibilityMatrix(w)
    }
    livingBins <- which(rowSums(w$bcm)>0 | colSums(w$bcm)>0)
    g <- w$bcm[livingBins,livingBins]
    if(sum(w$cm)==0 || length(g)<1){
        cat("empty compatibility matrix\n")
        return()
    }
    ## make the labels
    gRuleInv <- w$ruleInv[livingBins,] ## just the bins in g
    theseLabels <- paste("[",livingBins,"]",sep="")
    for(i in seq.int(nrow(w$ruleList))){ ## for each rule
        theseBins <- gRuleInv[,i]>0
        theseLabels[theseBins] <- paste(
            theseLabels[theseBins],
            "\n",w$ruleList[i,"inBall"],"->",w$ruleList[i,"outBall"],
            " (",gRuleInv[theseBins,i],")",
            sep=""   
        )
    }
    ## make layout coordinates:
    if(useGrid){
        theseCoords <- cbind(
            ((livingBins-1) %% 10),
            -(floor((livingBins-1)/10))
        )
    }else{
        theseCoords <- gplot.layout.kamadakawai(g,list())
    }
    gplot(
        as.matrix(g),
        label=theseLabels,
        boxed.labels=FALSE,
        label.pos=5,
        label.cex=label.cex,
        label.col="#FFFFFF",
        vertex.sides=4,
        vertex.rot=45,
        vertex.cex=vertex.cex,
        usecurve=FALSE,
        displayisolates=FALSE,
        coord=theseCoords,
        jitter=FALSE,
        ...
    )

}

plotWorldGrid <- function(w,
    boxed.labels=FALSE,label.pos=5,vertex.cex=3,label.cex=.65,jitter=TRUE,
    edge.col="#66553399",vertex.col="#FFFFFF33",vertex.border="#FFFFFF33",
    gridCol="#22334455",...
){
    ## first make sure we have cm and bcm, create mcm
    if(is.null(w$cm)) w$cm <- getCompatibilityMatrix(w)
    if(is.null(w$bcm)) w$bcm <- getBinCompatibilityMatrix(w)
    bcm <- w$bcm[-isolates(w$bcm),-isolates(w$bcm)]
    mcm <- w$bcm
    mcm[get.adjacency(agraph.moore(w$n))==0] <- 0
    mcm <- mcm[-isolates(w$bcm),-isolates(w$bcm)]
    cmBins <- as.numeric(rownames(w$cm))
    mcmBins <- as.numeric(rownames(mcm))
    ## and make a nice title
    title <- paste(
        paste(w$maxBallType,"ball",sep="-"),
        w$chemistry,
        if(w$env.endogenous){"endogenous"}else{"fixed"},
        if(w$env.rich){"rich"}else{"poor"},
        w$reproduction,
        paste(w$mixedMode*100,"% mixed",sep=""),
        paste("tick",w$tick),
        sep="; "
    )
    
    ########################
    ## First see if we can rotate the moore space to minimize edge wrap
    ########################
    ## put mcm on a grid to see if we can rotate it to avoid edge wrap
    gridDim <- sqrt(w$n)
    lyt <- cbind(bin=unique(cmBins),expand.grid(x=1:gridDim,y=1:gridDim)[unique(cmBins),])
    edgeRows <- row(mcm)[mcm>0]
    edgeCols <- col(mcm)[mcm>0]
    ## which moore x-coords do spatial edges cross
    xEdges <- cbind(lyt[rownames(mcm)[edgeRows],]$x,lyt[colnames(mcm)[edgeCols],]$x)
    xDiffs <- abs(apply(xEdges,1,diff))
    xBreaks <- if(any(xDiffs==1)){
        rev(table(c(
            apply(matrix(xEdges[xDiffs==1,],ncol=2),1,min),
            rep(gridDim,sum(xDiffs>1))
        ))) 
    }else{
        c("1"=0)
    }
    ## which moore y-coords do spatial edges cross?
    yEdges <- cbind(lyt[rownames(mcm)[edgeRows],]$y,lyt[colnames(mcm)[edgeCols],]$y)
    yDiffs <- abs(apply(yEdges,1,diff))
    yBreaks <- if(any(yDiffs==1)){
        rev(table(c(
            apply(matrix(yEdges[yDiffs==1,],ncol=2),1,min),
            rep(gridDim,sum(yDiffs>1))
        )))
    }else{
        c("1"=0)
    }
    ## Decide where to break the grid
    xb <- if(length(xBreaks)>=gridDim){
        as.numeric(names(xBreaks)[which.min(xBreaks)])
    }else{
        max(seq(gridDim)[-as.numeric(names(xBreaks))])
    }
    yb <- if(length(yBreaks)>=gridDim){
        as.numeric(names(yBreaks)[which.min(yBreaks)])
    }else{
        max(seq(gridDim)[-as.numeric(names(yBreaks))])
    }
    ## and now rotate lyt
    if(xb<10){lyt$x <- (lyt$x-xb-1) %% gridDim + 1}
    if(yb<10){lyt$y <- (lyt$y-yb-1) %% gridDim + 1}
    
    ########################
    ## Get first rule name for each bin
    ## (add subsequent rule names later)
    ########################
    ruleListLab <- apply(w$ruleList,1,paste,collapse="-")
    firstRules <- apply(w$ruleInv[lyt$bin,],1,function(x){which(x>0)[1]})
    ruleLab <- ruleListLab[firstRules]
    for(i in seq(length(ruleLab))){
        ruleLab[i] <- paste(ruleLab[i],"(",w$ruleInv[lyt$bin[i],firstRules[i]],")",sep="")
    }
    

    ########################
    ## Now find multi-rule bins and give each rule its own coordinates
    ## (while looping, fix rule names)
    ########################
    multirule <- unique(cmBins[duplicated(cmBins)])
    for(b in multirule){
        i <- which(lyt$bin==b)
        nRules <- sum(cmBins==b)
        xmax <- min(.25,nRules*.05)
        ymax <- min(.35,nRules*.1)
        newLyt <- cbind(
            bin=b,
            x=lyt$x[i] + seq(from=-xmax,to=xmax,length.out=nRules),
            y=lyt$y[i] + seq(from=ymax,to=-ymax,length.out=nRules)
        )
        rownames(newLyt) <- rep(b,nRules)
        lyt <- rbind(
            if(i>1){lyt[1:(i-1),]},
            newLyt,
            if(i<nrow(lyt)){lyt[(i+1):nrow(lyt),]}
        )
        ## and don't forget the rule labels
        newRuleLab <- paste(
            ruleListLab[w$ruleInv[b,]>0],
            "(",w$ruleInv[b,][w$ruleInv[b,]>0],")",sep=""
        )
        ruleLab <- c(
            if(i>1){ruleLab[1:(i-1)]},
            newRuleLab,
            if(i<length(ruleLab)){ruleLab[(i+1):length(ruleLab)]}
        )
    }

    ########################
    ## figure out which edges in cm are not spatially adjacent
    ########################
    ## slow but works:
    for(i in which(bcm-mcm>0)){
        from <- rownames(bcm)[row(bcm)[i]]
        to <- colnames(bcm)[col(bcm)[i]]
        x <- rownames(w$cm)==from
        y <- colnames(w$cm)==to
        w$cm[x,y] <- w$cm[x,y]/2
    }
    
    ########################
    ## Plot vertices and edges
    ########################
    if(jitter){
        lyt$x <- jitter(lyt$x,amount=.05)
        lyt$y <- jitter(lyt$y,amount=.05)
    }
    gplot(
        w$cm,coord=lyt[c("x","y")],edge.lty=w$cm,label=ruleLab,main=title,
        edge.col=edge.col,boxed.labels=boxed.labels,label.pos=label.pos,label.cex=label.cex,
        vertex.col=vertex.col,vertex.border=vertex.border,vertex.cex=vertex.cex,
        sub=if(worldIsAlive(w)){"Alive"}else{"Dead"},
        ...
    )
    
    ########################
    ## and finally, draw the grid
    ########################
    xgrid <- seq(round(min(lyt$x)),round(max(lyt$x)))
    ygrid <- seq(round(min(lyt$y)),round(max(lyt$y)))
    for(x in xgrid-0.5){
        lines(c(x,x),c(min(ygrid)-0.5,max(ygrid)+0.5),col=gridCol)
    }
    lines(c(x,x) + 1,c(min(ygrid)-0.5,max(ygrid)+0.5),col=gridCol)
    for(y in ygrid-0.5){
        lines(c(min(xgrid)-0.5,max(xgrid)+0.5),c(y,y),col=gridCol)
    }
    lines(c(min(xgrid)-0.5,max(xgrid)+0.5),c(y,y) + 1,col=gridCol)
    
}


###########################
## Statistics gatherers
##  The following few functions gather statistics from model objects.
##  The code is a little spaghetti, but it works well as is.
###########################

## Creates a data.frame with some summary stats the state of the world.
## This is the basic stats-gathering function, called heavily by the others.
getWorldStats <- function(
    w,countNonrecipEdges=TRUE,runId=1,maxCycleLength=9,countSubsystemsOnDeadWorld=TRUE,
    uniqueItem="rule",allowSelfPass=FALSE,...
){
    ## refresh compatibility matrix even if it exists... just for certainty
    cm <- w$cm <- getCompatibilityMatrix(w)
    
    ## what are the actual rule indices of the nodes?
    nodeRuleInd <- unlist(lapply(as.numeric(unique(rownames(cm))),function(n){
        which(w$ruleInv[n,]>0)
    }))
    
    if(countNonrecipEdges){
        if(!is.null(dim(cm))){
            nrcm <- cm - symmetrize(w$cm,"strong")
        }else{
            nrcm <- cm
        }
        nrNodeRuleInd <- unlist(lapply(as.numeric(unique(rownames(nrcm))),function(n){
            which(w$ruleInv[n,]>0)
        }))
    }
    maxK <- min(nrow(cm),max(w$maxBallType,maxCycleLength))
    if(!is.null(dim(cm))){
        kcc <- kcycle.census(cm,maxlen=maxK,
            tabulate.by.vertex=TRUE,
            cycle.comembership="sum"
        )
        k <- which(kcc$cycle.count[,"Agg"]>0)+1
        #cat("identifying cycles of lengths (",
        #    paste(k,collapse=","),") on a ",
        #    nrow(cm),"-node graph.\n",sep=""
        #)
        cycles <- identifyCycles(cm,k,...)
        cycleLengths <- lapply(cycles,length)
        if(countNonrecipEdges){
            nrCycles <- identifyCycles(nrcm,k,...)
            nrCycleLengths <- lapply(nrCycles,length)
            if(length(nrCycleLengths)<1){
                nrCycleLengths <- list(-1) ## hack
            }
        }
        
        
        ## Figure out which cycles are valid by not repeating a `uniqueItem` (arg)
        ## first for all cycles
        if(length(k)>0){
            ## this code in here is a little wonky. After all the 'itemCycles' and
            ## 'itemsUnique' stuff was in there, I needed to add multiple parasite counters
            ## in each run. Thus we end up with all three of ruleCycles, binCycles and
            ## itemCycles. Not ideal but functional.
            binCycles <- lapply(cycles,function(l){
                as.numeric(rownames(cm)[l])
            })
            if(!allowSelfPass){
                ## now that we have binCycles we can cull out any cycles that
                ## have a self-pass in them.
                ## who has them?
                hasSelfPass <- sapply(binCycles,function(x){
                    any(diff(x)==0) || (length(x)>1 && x[1]==x[length(x)])
                })
                ## get rid of 'em
                cycles <- cycles[!hasSelfPass]
                cycleLengths <- cycleLengths[!hasSelfPass]
                binCycles <- binCycles[!hasSelfPass]
            }
            binsUnique <- unlist(lapply(binCycles,function(x){!any(duplicated(x))}))
            ruleCycles <- lapply(cycles,function(l){
                nodeRuleInd[l]
            })
            rulesUnique <- unlist(lapply(ruleCycles,function(x){!any(duplicated(x))}))
            ruleBinCycles <- mapply(paste,ruleCycles,binCycles)
            ruleBinsUnique <- unlist(lapply(ruleBinCycles,function(x){!any(duplicated(x))}))
            if(uniqueItem=="rule"){
                itemCycles <- ruleCycles
                itemsUnique <- rulesUnique
            }else if(uniqueItem=="bin"){
                itemCycles <- binCycles
                itemsUnique <- binsUnique
            }else{ # uniqueItem=="ruleBin", probably
                itemCycles <- ruleBinCycles
                itemsUnique <- ruleBinsUnique
            }
            if(any(itemsUnique)){
                kUniqueItems <- sort(unique(unlist(cycleLengths[itemsUnique])))
            }else{
                kUniqueItems <- numeric()
            }
        }else{
            itemCycles <- ruleCycles <- binCycles <- list()
            itemsUnique <- rulesUnique <- binsUnique <- logical()
            kUniqueItems <- numeric()
        }
        
        ## now for nonreciprocal cycles
        if(countNonrecipEdges && length(nrCycles)>0){
            nrBinCycles <- lapply(nrCycles,function(l){
                as.numeric(rownames(nrcm)[l])
            })
            if(!allowSelfPass){
                ## now that we have nrBinCycles we can cull out any cycles that
                ## have a self-pass in them.
                ## who has them?
                hasSelfPass <- sapply(nrBinCycles,function(x){
                    any(diff(x)==0) || (length(x)>1 && x[1]==x[length(x)])
                })
                ## get rid of 'em
                nrCycles <- nrCycles[!hasSelfPass]
                nrCycleLengths <- nrCycleLengths[!hasSelfPass]
                nrBinCycles <- nrBinCycles[!hasSelfPass]
            }
            nrBinsUnique <- unlist(lapply(nrBinCycles,function(x){!any(duplicated(x))}))
            nrRuleCycles <- lapply(nrCycles,function(l){
                nrNodeRuleInd[l]
            })
            nrRulesUnique <- unlist(lapply(nrRuleCycles,function(x){!any(duplicated(x))}))
            
            
            
            if(uniqueItem=="rule"){
                nrItemCycles <- nrRuleCycles
                nrItemsUnique <- nrRulesUnique
            }else{ # uniqueItem=="bin"
                nrItemCycles <- nrBinCycles
                nrItemsUnique <- nrBinsUnique
            }
            if(any(nrItemsUnique)){
                kNrUniqueItems <- sort(unique(unlist(nrCycleLengths[nrItemsUnique])))
            }else{
                kNrUniqueItems <- numeric()
            }
            nrItemCycleLengths <- lapply(nrItemCycles,length)
        }else{
            nrItemCycles <- nrRuleCylces <- nrBinCycles <- list()
            nrItemsUnique <- nrRulesUnique <- nrBinsUnique <- logical()
            kNrUniqueItems <- numeric()
        }
        if(length(nrItemCycles)>0){
            nrItemCycleLengths <- lapply(nrItemCycles,length)
        }else{
            nrItemCycleLengths <- list(-1) # hack again
        }
        
        
        ## this bit finds all the edges indicated in `cycles`, then
        ## creates an adjacency matrix from cm without these edges
        cycleEdges.unrestricted <- unique(unlist(lapply(cycles,function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
        cycleEdges.uniqueRules <- unique(unlist(lapply(cycles[rulesUnique],function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
        cycleEdges.uniqueBins <- unique(unlist(lapply(cycles[binsUnique],function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
        ## nonreciprocal version of above:
        nrCycleEdges.unrestricted <- unique(unlist(lapply(nrCycles,function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
        nrCycleEdges.uniqueRules <- unique(unlist(lapply(nrCycles[nrRulesUnique],function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
        nrCycleEdges.uniqueBins <- unique(unlist(lapply(nrCycles[nrBinsUnique],function(l){
            res <- list(c(l[length(l)],l[1]))
            for(i in seq.int(1,length(l)-1)){
                res <- c(res,list(l[c(i,i+1)]))
            }
            return(res)
        }),recursive=FALSE))
    
        nonCycleCm.uniqueRules <- nonCycleCm.uniqueBins <- cm
        for(l in cycleEdges.uniqueRules){
            nonCycleCm.uniqueRules[l[1],l[2]] <- 0
        }
        for(l in cycleEdges.uniqueBins){
            nonCycleCm.uniqueBins[l[1],l[2]] <- 0
        }
        ## nonreciprocal version of above (plus unrestricted version):
        nrNonCycleCm.unrestricted <-nrNonCycleCm.uniqueRules <- nrNonCycleCm.uniqueBins <- cm
        for(l in nrCycleEdges.unrestricted){
            nrNonCycleCm.unrestricted[l[1],l[2]] <- 0
        }
        for(l in nrCycleEdges.uniqueRules){
            nrNonCycleCm.uniqueRules[l[1],l[2]] <- 0
        }
        for(l in nrCycleEdges.uniqueBins){
            nrNonCycleCm.uniqueBins[l[1],l[2]] <- 0
        }
    }else{
        cycles <- nrCycles <- list()
        kUniqueItems <- numeric()
        nrItemCycles <- list()
        kNrUniqueItems <- numeric()
        cycleEdges.unrestricted <- cycleEdges.uniqueRules <- cycleEdges.uniqueBins <- list()
        nonCycleCm.uniqueRules <- nonCycleCm.uniqueBins <- cm
        itemCycles <- ruleCycles <- binCycles <- list()
        itemsUnique <- rulesUnique <- binsUnique <- logical()
        nrNonCycleCm.unrestricted <-nrNonCycleCm.uniqueRules <- nrNonCycleCm.uniqueBins <- cm
        nrItemCycles <- nrRuleCylces <- nrBinCycles <- list()
        nrItemsUnique <- nrRulesUnique <- nrBinsUnique <- logical()
    }
    
    
    
    
    
    thisWorldIsAlive.noSP <- worldIsAlive(w,removeSelfPass=TRUE)
    thisWorldIsAlive <- worldIsAlive(w,removeSelfPass=FALSE)
    
    ## a couple more calculations for parasite-finding:
    allNodes <- seq.int(nrow(cm))
    uniqueRuleNodes <- unique(unlist(cycles[rulesUnique]))
    uniqueBinNodes <- unique(unlist(cycles[binsUnique]))
    nrUniqueRuleNodes <- unique(unlist(nrCycles[nrRulesUnique]))
    nrUniqueBinNodes <- unique(unlist(nrCycles[nrBinsUnique]))
    
    ## make the run-level results data frame
    res <- cbind(
        ## same for whole run:
        runId=runId,
        as.data.frame(w[names(hyper3params())[names(hyper3params()) %in% names(w)]]),
        #searchType=if(is.null(w$searchType)){NA}else{w$searchType},
        isAlive=thisWorldIsAlive,
        isAlive3Plus=thisWorldIsAlive & any(kUniqueItems>2), # 'complex' survival
        isAliveNr3Plus=thisWorldIsAlive & length(nrItemCycles)>0 & any(kNrUniqueItems>2),
        isAlive.removeSelfPass=thisWorldIsAlive.noSP,
        isAlive3Plus.removeSelfPass=thisWorldIsAlive.noSP & any(kUniqueItems>2), # 'complex' survival
        isAliveNr3Plus.removeSelfPass=thisWorldIsAlive.noSP & length(nrItemCycles)>0 & any(kNrUniqueItems>2),
        livingRuleCount=sum(w$ruleInv>0),
        distinctRuleCount=sum(colSums(w$ruleInv)>0),
        livingBinCount=if(!is.null(w$deadBins)){sum(!w$deadBins)}else{sum(w$isAlive)},
        parasiteEdgeCount.allCycles=parasiteEdge.count(cm,cc=kcc),
        parasiteEdgeCount.uniqueRules=parasiteEdge.count(nonCycleCm.uniqueRules),
        parasiteEdgeCount.uniqueBins=parasiteEdge.count(nonCycleCm.uniqueBins),
        parasiteRuleCount.allCycles=parasite.count(cm,cc=kcc),
        parasiteRuleCount.uniqueRules=sum(!(allNodes %in% uniqueRuleNodes)),
        parasiteRuleCount.uniqueBins=sum(!(allNodes %in% uniqueBinNodes)),
        nonrecipParasiteEdgeCount.allCycles=parasiteEdge.count(nrNonCycleCm.unrestricted),
        nonrecipParasiteEdgeCount.uniqueRules=parasiteEdge.count(nrNonCycleCm.uniqueRules),
        nonrecipParasiteEdgeCount.uniqueBins=parasiteEdge.count(nrNonCycleCm.uniqueBins),
        nonrecipParasiteRuleCount.allCycles=parasite.count(nrNonCycleCm.unrestricted),
        nonrecipParasiteRuleCount.uniqueRules=sum(!(allNodes %in% nrUniqueRuleNodes)),
        nonrecipParasiteRuleCount.uniqueBins=sum(!(allNodes %in% nrUniqueBinNodes))
    )
    ## add some code to get "islands" under mixed-mode
    if("mixedMode" %in% names(w)){
        if(!is.null(dim(cm))){
            mcm <- w$bcm
            mcm[get.adjacency(agraph.moore(w$n))==0] <- 0
            mcm <- mcm[-isolates(w$bcm),-isolates(w$bcm)]
            
            
            ## some plots
            if(TRUE){
                title <- paste(
                    paste(w$maxBallType,"ball",sep="-"),
                    w$chemistry,
                    if(w$env.endogenous){"endogenous"}else{"fixed"},
                    if(w$env.rich){"rich"}else{"poor"},
                    w$reproduction,
                    paste(w$mixedMode*100,"% mixed",sep=""),
                    sep="; "
                )
                mcmc <- w$bcm
                mcmc[get.adjacency(agraph.moore(w$n))==0 & w$bcm==1] <- .5
                pdf(file=paste(runId,".islands.pdf",sep=""),width=8,height=8)
                gplot(mcmc,coord=expand.grid(1:10,1:10),jitter=F,edge.lty=mcmc,main=title)
                dev.off()
                pdf(file=paste(runId,".rules.pdf",sep=""),width=8,height=8)
                plotWorldRules(w)
                dev.off()
            }
            
            
            compd.strong <- component.dist(mcm,connected="strong")$csize
            compd.weak <- component.dist(mcm,connected="weak")$csize
            res <- cbind(res,data.frame(
                mixedMode=w$mixedMode,
                mooreIslandCount.strong=length(compd.strong),
                mooreIslandNonsingularCount.strong=sum(compd.strong>1),
                mooreIslandCount.weak=length(compd.weak),
                mooreIslandNonsingularCount.weak=sum(compd.weak>1)
            ))
        }else{
            res <- cbind(res,data.frame(
                mixedMode=w$mixedMode,
                mooreIslandCount.strong=NA,
                mooreIslandNonsingularCount.strong=NA,
                mooreIslandCount.weak=NA,
                mooreIslandNonsingularCount.weak=NA
            ))
        }
    }
    
    #if(is.na(res$isAliveNr3Plus)){recover()}
    if(!is.null(dim(cm)) && length(k)>0){
        ## which cycles are subsystems?:
        if(thisWorldIsAlive || countSubsystemsOnDeadWorld){
            ## first orient all cycles in common
            cycles <- lapply(cycles,function(x){
                ## first translate nodes to rules
                x <- nodeRuleInd[x]
                ## now standardize their order
                cl <- length(x)
                maxRule <- max(x)
                minInd <- which(x==min(x))
                ## Derive some scores that will reliably differentiate
                ## rotations that start with the same rule. This treats
                ## them as the digits of a base-maxRule number.
                scores <- unlist(lapply(minInd,function(i){
                    if(i>1){
                        x <- x[c(i:cl,1:(i-1))]
                    }
                    sum((x-1)*maxRule^(cl:1))
                }))
                minInd <- minInd[scores==min(scores)][1]
                ## got it. now do the rotation
                if(minInd>1){
                    x <- x[c(minInd:cl,1:(minInd-1))]
                }
                return(x)
            })
            
            ## now get all that are subsystems and are unique in `uniqueItem` argument
            if(length(cycles)>0){
                subsystems <- cycles[!duplicated(cycles) & itemsUnique]
            }else{
                subsystems <- list()
            }
            subsystemLengths <- lapply(subsystems,length)
        }else{
            subsystems <- list()
            subsystemLengths <- list()
        }
        ## add the stats:
        res <- cbind(
            res[rep(1,length(k)),],
            ## for each cycle length:
            cycleLength=k,
            rawCycleCount=kcc$cycle.count[,"Agg"][as.character(k)],
            cycleCount=if(length(cycleLengths)>0){
                    tabulate(unlist(cycleLengths))[k]
                }else{
                    0
                },
            nonRecipCycleCount=if(length(nrCycleLengths)>0){
                        tabulate(unlist(nrCycleLengths))[k]
                    }else{
                        0
                    },
            uniqueItemCycleCount=if(any(itemsUnique)){
                tabulate(unlist(cycleLengths[itemsUnique]),nbins=max(k))[k]
            }else{
                0
            },
            uniqueNonRecipCycleCount=tabulate(unlist(nrItemCycleLengths),nbins=max(k))[k],
            subsystemCount=unlist(lapply(k,function(thisK){
                sum(subsystemLengths==thisK)
            })) 
        )
        
    }else{
        res <- cbind(
            res,
            cycleLength=NA,
            rawCycleCount=NA,
            cycleCount=NA,
            nonRecipCycleCount=NA,
            uniqueItemCycleCount=NA,
            uniqueNonRecipCycleCount=NA,
            subsystemCount=NA
        )
    }
    return(res)
}

# get stats on a list of hyper3world objects, serially
getBatchWorldStats <- function(wList,...){
    ## would use mapply, but it returns a list or matrix, not a data frame.
    ## also, this gives us a progress bar :)
    cat("getting stats on",length(wList),"worlds:")
    flush.console()
    pb <- txtProgressBar(min=0,max=length(wList),style=3)
    res <- data.frame()
    cat("\n")
    for(i in seq.int(length(wList))){
        res <- rbind(res,getWorldStats(wList[[i]],runId=i,...))
        setTxtProgressBar(pb,i)
        gc()
    }
    close(pb)
    return(res)
}

# same as getBatchWorldStats, but uses multiple cores if possible
getmcBatchWorldStats <- function(wList,numProcs=getOption("cores"),...){
    require(multicore)
    if(is.null(numProcs)) numProcs <- multicore:::volatile$detectedCores
    cat("getting stats on",numWorlds <- length(wList),"worlds:\n")
    flush.console()
    worldNum <- 0
    mclapply(wList,function(w,...){
        worldNum <<- worldNum + 1
        cat(Sys.getpid(),"run",worldNum,"of about",numWorlds/numProcs,"\n")
        getWorldStats(w,runId=as.numeric(paste(Sys.getpid(),worldNum,sep=".")),...)
    },mc.cores=numProcs,...)
}

# call getBatchWorldStats on save()ed lists of hyper3world objects
getFileBatchWorldStats <- function(fileList,resList,...){
    ## give a character vector of files containing data
    ## resList is a list of what the res objects are called
    res <- data.frame()
    
    ## create an environment to avoid stepping on toes (eg "res" %in% resList)
    resEnv <- new.env()
    maxId <- 0
    for(i in 1:length(fileList)){
        cat("\nimporting",fileList[i])
        load(fileList[i],env=resEnv)
        cat("\nanalyzing",resList[i],"\n")
        flush.console()
        newRes <- getBatchWorldStats(get(resList[i],env=resEnv),...)
        newRes$runId <- newRes$runId + maxId
        if(length(names(res))>1){
            newNames <- intersect(names(res),names(newRes))
            res <- rbind(res[newNames],newRes[newNames])
        }else{
            res <- rbind(res,newRes)
        }
        maxId <- max(res$runId)
    }
    return(res)
}

# same as getFileBatchWorldStats, but uses multiple cores if possible
getmcFileBatchWorldStats <- function(fileList,resList,numProcs=getOption("cores"),...){
    ## give a character vector of files containing data
    ## resList is a list of what the res objects are called
    res <- NULL
    
    ## create an environment to avoid stepping on toes (eg "res" %in% resList)
    resEnv <- new.env()
    maxId <- 0
    for(i in 1:length(fileList)){
        cat("\nimporting",fileList[i])
        load(fileList[i],env=resEnv)
        cat("\nanalyzing",resList[i],"\n")
        flush.console()
        newRes <- getmcBatchWorldStats(get(resList[i],env=resEnv),numProcs=numProcs,...)
        if(!is.data.frame(newRes)){newRes <- do.call(rbind,newRes)}
        newRes$runId <- newRes$runId + maxId
        if(!is.null(res)){
            newNames <- intersect(names(res),names(newRes))
            res <- rbind(res[newNames],newRes[newNames])
        }else{
            res <- newRes
        }
        maxId <- max(res$runId)
        rm(newRes);gc()
    }
    return(res)
}





## identifyCycles returns a list of the cycles in an adjacency matrix,
## where a cycle is identified by a vector of vertex indices.
identifyCycles <- function(
    am,k,
    remove.symmetric=FALSE,allow.repeat.vertices=FALSE,maxPaths=100000
){
    ## if we're not looking for cycles of any length, return empty list
    if(length(k)<1) return(list())
    
    ## ignore self-loops -- we don't care about 1-cycles.
    diag(am) <- 0
    
    ## if we're ignoring symmetric loops and every edge is bi-directional and k>2,
    ## then return an empty list. This is just here to speed things up in some cases
    if(remove.symmetric && all(am==t(am)) && any(k>2)) return(list()) ## heuristic
    
    ## initialize the "working path list" to be a 1-path for each vertex.
    pathList <- as.list(seq.int(nrow(am)))
    ## initialize results list as empty
    res <- list()
    ## as long as we have any paths in the working list, execute path-expansion
    while(length(pathList)>0){
        ## This is just here to deal with runaway paths if graph is too dense or large
        ## for this algorithm. It can be bypassed by setting maxPaths=Inf.
        if(length(pathList)>maxPaths){
            warning(
                "Path list surpassed maximum length of ",maxPaths,
                ". Abandoning identifyCycles, some will likely be missing.\n",
                "Ignoring this warning is probably a bad idea."
            )
            break
        }
        
        ## snag the first path in the list (FIFO / breadth-first)
        thisPath <- pathList[[1]]
        #cat(thisPath,"\n")
        
        ## if thisPath is of one of the lengths we're looking for, check if it's a cycle
        if(length(thisPath) %in% k){
            if(am[thisPath[length(thisPath)],thisPath[1]]>0){
                ## success! Path is a cycle, so add it to the results list
                res <- c(res,list(thisPath))
            }
        }
        
        ## now we find all the relevant paths that extend this one
        ## first make sure there is any k larger than the current path length
        if(length(thisPath) < max(k)){
            ## extension candidates:
            newNeighbors <- which(am[thisPath[length(thisPath)],]>0)
            ## make sure they're not repeats
            if(!allow.repeat.vertices){
                newNeighbors <- newNeighbors[!(newNeighbors %in% thisPath)]
            }
            ## add the remaining neighbors to the end of the working path list
            for(i in newNeighbors) pathList <- c(pathList,list(c(thisPath,i)))
        }
        
        ## be sure to pop off the path we've been working with (could do this above).
        pathList <- pathList[-1]
    }
    ## just need to orient all in common and get rid of repeats
    ## (currently only fully works for allow.repeat.vertices==FALSE)
    res <- lapply(res,function(l){
        minInd <- which(l==min(l))[1]
        if(minInd>1) l <- l[c(minInd:length(l),1:(minInd-1))]
        l
    })
    if(remove.symmetric && any(k>2)){
        revRes <- lapply(res,function(l){
            l <- rev(l)
            minInd <- which(l==min(l))
            if(minInd>1) l <- l[c(minInd:length(l),1:(minInd-1))]
            l
        })
        res <- res[!(res %in% revRes)]
    }
    
    return(unique(res))
}
parasite.count <- function(g,cc=NULL){
    ## this function finds nodes that are not in a cycle
    ## but are connected to nodes in a cycle. optional arg
    ## `cc` is the output from kcycle.census() in sna.
    ## `g` should be an adjacency matrix.
    if(length(g)<=1) return(0)
    if(is.null(cc)) cc <- kcycle.census(g,max(2,nrow(g)))
    ## which aren't in a cycle
    noCycle <- which(apply(cc$cycle.count,2,sum)[-1]==0)
    res <- 0
    r <- reachability(g)
    for(i in noCycle){
        if(any(!(which(r[i,]+r[,i]>0) %in% noCycle))){
            res <- res + 1
        }
    }
    return(res)
}
parasiteEdge.count <- function(g,cc=NULL){
    ## this function counts edges that are not a part of any cycle
    if(length(g)<=1) return(0)
    if(is.null(cc)) cc <- kcycle.census(g,max(2,nrow(g)),cycle.comembership="sum")
    return(sum(cc$cycle.comemb[g>0]==0))
}

